package PureMessage::Manager::Response::Policy;
use PureMessage::Sieve qw(sieve_parse_file sieve_print sieve_node sieve_text);
use PureMessage::Manager::Utils qw (html_escape);
use PureMessage::Version qw($PREFIX);
use PureMessage::Manager::Policy qw(MIDDLE_COL RIGHT_COL add_cmd_before);
use strict;
use PureMessage::Manager::Request;
use PureMessage::Manager::Widget::ToolBar;
use PureMessage::Manager::Widget::Dialog;
use PureMessage::Manager::Widget;
use HTML::Entities;

our  @ISA = ('PureMessage::Manager::Response');

# use globals for the prototype, must be rewritten for the real thing
my $indent = 0;
my $ruleset_id = 0;
my $opened_level;
my $kw;
my $class;
my $max_args = 10;
my $tests_help = "/Help/mgr/mgr_pol_tests.html";
my $actions_help = "/Help/mgr/mgr_pol_actions.html";
our $rule_style = 'black';

sub new {
    my $proto = shift;
    my $self = $proto->SUPER::new(@_);
    $self->{'html_meta_charset'} = 'utf-8';
    $self->addRestartDlg();
    return $self;
}

sub init_file {
    my $self = shift;
    my %opts = @_;
    $self->{file} = $opts{file} || "$PREFIX/etc/policy.siv";
    $self->{clp_file} = $opts{clp_file} || "$PREFIX/etc/.clipboard.siv";
    unless (-f $self->{file}) {
        if (open (my $f, ">$self->{file}")) {
            $self->raise_error("File $self->{file} is missing. An empty file was created.");
        }
        else {
            $self->raise_error("Fatal error: can't create policy script file $self->{file}");
        }
    }
    if ( -f $self->{file}.".edit" ) {
        if ( $self->{REQUEST}->query('revert') ) {
            unlink $self->{file}.".edit";
            # XXX error check
            $self->print_notification("Reverted to the original policy");
        }
        elsif ( $self->{REQUEST}->query('commit') ) {
            # XXX use File::Atomic here (?)
            my $check = `$PREFIX/bin/pmx-policy check --script $self->{file}.edit 2>&1`;
            if ($? == 0) {
                if (rename "$self->{file}.edit", $self->{file}) {
                    # XXX add this notification to RestartDlg:
                    $self->print_notification("The changes have been committed.");
                    $self->addRestartDlg('query_ok'=>1);
                }
                else {
                    $self->raise_error("Error: unable to commit changes: $self->{file}: $!");
                    $self->commit_dlg();
                }
            }
            else {
                $self->raise_error("Error: $check");
                $self->raise_error('Syntax errors found - changes not committed.  Click <a href="?revert=1">revert</a> to go back to the last saved policy.');
                $self->commit_dlg();
            }
        }
        else {
            $self->{file} .= ".edit";
            $self->commit_dlg();
        }
    }
}

sub save_source {
    my $self = shift;
    my $source = shift;
    $self->init_file() unless $self->{file};
    
    $self->{file} .= ".edit" unless $self->{file} =~ /\.edit/;
    open ( my $S, ">$self->{file}" ) 
        || return $self->raise_error("Can't open file for writing $self->{file}.edit");
    my $src = $source || $self->{REQUEST}->query('sieve');
    $src =~ s/\r//g;
    print $S $src;
    $self->print_notification("Your changes have been saved.");
    $self->commit_dlg();
}

sub format_source {
  my ($self,$filename) = @_;
  open(F, $filename);
  my @code = <F>;
  my $html = join("", @code);
  $html =~ s/\t/&nbsp;&nbsp;&nbsp;&nbsp;/g;
  $html =~ s/ /&nbsp;/g;
  $html =~ s/(if|else|elsif|{|})(\W)/<font color='#A21E1E'><b>$1<\/b><\/font>$2/g;
  $html =~ s/([\n\s]#.*)/<font color='#808080'>$1<\/font>/g;
  $html =~ s/\"(.*)\"/"<font color='#7b007b'>$1<\/font>"/g;
  $html =~ s/\r?\n/<br>/g;
  return <<EOF;
        <span style='font-family: Lucidatypewriter, "Lucida console", Courier'>$html</pre>
EOF

 
}

sub show_source {
    my $self = shift;
    $self->init_file() unless $self->{file};
    my $html = $self->format_source($self->{file}); 
    return <<EOT;
    <br>
    <table cellpadding=4 cellspacing=0 width="100%">
    <tr><td>
        Source code for the policy script (Sieve)<br>
        File: <a href="index.cgi?editsource=1">$self->{file}</a>
    </td></tr>
    <tr><td>
      $html
    </td></tr>
    </table>
EOT
}

sub edit_source {
    my $self = shift;
    $self->init_file() unless $self->{file};
    open(F, $self->{file});
    my @code = <F>;
    my $html = join("", @code);
    
    return <<EOT;
    <form action="index.cgi" method=POST>
    <table cellpadding=4 width="100%">
    <tr><td class="light">
        Source code for the policy script (Sieve)<br>
        File: $self->{file}
    </td></tr>
    <tr><td>
<textarea name=sieve cols=68 rows=25 style='width: 100%; font-family: Fixed, Courier, Fixed; font-size: 10pt; border-style: none; '>$html
</textarea>
    </td></tr>
    <tr><td>
        <input type=submit name="savesource" value="Save" class=button>
    </td></tr>
    </table>
    </form>
EOT
}

sub show_diff {
    my $self = shift;
    $self->init_file() unless $self->{file};
    my $file = $self->{file};
    $file =~ s/\.edit$//;
    if ( $file eq $self->{file} ) {
        return "There were no uncommitted changes made to the policy.";
    }
    else {
        open(F, "diff $file $self->{file}|");
        my @code = <F>;
        my $html = "<b>Changes:</b><br>";
        foreach (@code) {
            my $color = /^\</ ? "#A21E1E" : ( /^\>/ ? "#127719" : "#000000" );
            s/\t/&nbsp;&nbsp;&nbsp;&nbsp;/g;
            s/ /&nbsp;/g;
            $html .= "<font color='$color'>$_</font>";
        }
        $html =~ s/\r?\n/<br>/g;
		
        return "<span style='font-family: Lucidatypewriter, \"Lucida console\", Courier'>$html</pre>";
    }
}

sub show_rules {
    my $self = shift;
    $self->init_file() unless $self->{file};
    $self->{id} = $self->{REQUEST}->query('id');
    if ( $self->{REQUEST}->query('modified') 
        and not $self->{REQUEST}->query('delete') ) {
        $self->{save} = 1;
    }
    if ( $self->{REQUEST}->query('add_else') ) {
        $self->{add_else} = $self->{id};
        $self->{id} = undef;
    }
    elsif ( $self->{REQUEST}->query('add_if') ) {
        $self->{add_if} = defined $self->{id} ? $self->{id} : -1;
        $self->{id} = undef;
    }
    if ( $self->{REQUEST}->query('paste') ) {
        $self->{paste} = $self->{id};
        $self->{id} = undef;
    }
    eval {
        $self->{script} = sieve_parse_file($self->{file});
        $self->{p} = PureMessage::Manager::Policy->new();
        $self->{_actions} = {};
        foreach ($self->{p}->action_list) {
            $self->{_actions}{$_->name} = $_;
        }
        $self->{_tests} = {};
        foreach ($self->{p}->test_list) {
            $self->{_tests}{$_->name} = $_;
        }
        $self->{tree_html} .= "\n<SCRIPT>\n".$self->javascript()."\n</SCRIPT>\n";
        $self->{tree_html} .= "<form action=\"index.cgi\" method=POST>";
       # $self->{tree_html} .= qq|<a id='clips_id' href='/Policy/clips.cgi' onClick="window.open(this.href, 'popup', 'width=300, height=400, toolbar=no, resizable=yes, scrollbars=yes'); return false;">Clip Repository</a>|;
        
        my $i = 0;
        while ( my $cmd = ($self->{script}->commands)[$i] ) {
            $self->ruleset($cmd, $self->{script}, $i);
            # increment the counter unless current element has been deleted from the list
            $i++ unless ($self->{script}->commands)[$i] != $cmd; 
        }
        if ( $self->{add_if} == -1 ) {
            my $cmd = $self->add_if($self->{script});
            $self->ruleset($cmd, $self->{script});
        }
        $self->{tree_html} .= "</form>";
    };
    if ( $@ ) {
        $self->raise_error("Can't read modified policy script: $@ $!");
        return;
    }
    $self->{tree_html} .= "<br><a href='?add_if=1'>add main rule</a>";
    $self->{tree_html};
}

sub commit_dlg {
    my $self = shift;
    $self->{commit_dlg} = <<EOT;
<table cellpadding=4 cellspacing=0 width="100%">
<tr>
<td width="1" class="infoBox">
    <img src="/img/info.gif">
</td>
<td class="infoBox"><b>You have uncommitted changes.</b>
</td>
<td align="right" class="infoBox">
<a href="?commit=1">Commit</a> | 
<a href="?revert=1">Revert</a> | 
<a href="?diff=1">Source Diff</a>
</td>
</tr></table>
EOT
}

sub rule {
    my($self, $rule, $arg) = @_;
    if (length $rule->name) {
        $self->{tree_html} .= <<EOT;
<table cellspacing=0 border=0 width='100%'>
<tr><td $class valign=top width='1'>
EOT
        if ( $kw eq "if" ) { 
            if ( $indent == 0 ) {
                $self->{tree_html} .= "<image src='/img/top_if.gif'>";
            }
            else {
                $self->{tree_html} .= "<image src='/img/if.gif'>";
            }
        }
        elsif ( $kw eq "new" ) { 
            $self->{tree_html} .= "<image src='/img/top_if.gif'>";
        }
        else {
            $self->{tree_html} .= "<image src='/img/or.gif'>";
        }
        $self->{tree_html} .= "</td>";
        if ( $rule->{new} or $rule->{id} eq $self->{id} ) {
            $self->{tree_html} .= "<td width=\"100%\">";
            $self->{tree_html} .= $self->ThemeSectionHeader(width=>"100%")->output();
            $self->{tree_html} .= $self->edit_rule($rule);
            $class = "class=tableRowAlt";
        }
        else {
            $self->{tree_html} .= "<td $class>";
            $self->{tree_html} .= "<img src='/img/x.gif' height=18 width=1>";
            $self->{tree_html} .= "<a class=$rule_style href='?id=$rule->{id}'>$rule->{name}</a>";
        }
        $indent++;
        my $i = 0;
        while ( my $c = ($arg->commands)[$i] ) {
            $self->ruleset($c, $arg, $i);
            # increment the counter unless current element has been deleted from the list
            $i++ unless ($arg->commands)[$i] != $c; 
        }
        $indent--;
        if ($rule->{id} eq $self->{id}){
            $class = "";
            $self->{tree_html} .= $self->ThemeSectionFooter(toolbar=>$self->rule_tb($rule))->output();
        }
        $self->{tree_html} .= "</td></tr>";
        $self->{tree_html} .= "</table>\n";
    }
}

sub ruleset {
    my($self, $cmd, $parent, $pos) = @_;
    return if ($cmd->name ne "if");

    $kw = "if";
    my $i = 0;
    for my $arg ($cmd->args) {
        if ( $self->{add_if} eq $ruleset_id ) {
            $self->add_if($arg);
        }
        my $rule = $self->{p}->rule($arg, pos=>[$cmd,$i]);
        $kw = "else" if not $arg->test;
        $rule->{kw} = $kw;
        $rule->{parent} = $parent;
        $rule->{id} = $ruleset_id++;
        $self->{id} = $rule->{id} if $self->{new} and $self->{new} == $cmd;
        $i++;
        my $name = $rule->name();
        my $msg = "The rule <b>\"$name\"</b> has been ";
        if ( ($self->{REQUEST}->query('copy') or $self->{REQUEST}->query('cut')) and $rule->{id} eq $self->{id} ) {
            my $rule_node = $rule->get_node();
            unless  ($rule_node->test()) {
                $rule_node->test(sieve_node( "test", "true" ));
            }
            my $node = sieve_node( "if", "if", $rule_node );
            open (my $CLP, ">" . $self->{clp_file}) || $self->raise_error("Can't open clipboard file for writing");
            sieve_print($node, $CLP);
            if ( $self->{REQUEST}->query('copy') ) {
                $self->print_notification($msg . " copied to the clipboard.");
            }
            close ($CLP);
        }
        if ( ($self->{REQUEST}->query('delete') or $self->{REQUEST}->query('cut')) and $rule->{id} eq $self->{id} ) {
            # Delete rule
            eval {
                my @args = $cmd->args;
                if ( scalar @args > 1 ) {
                    # if there are two conds left
                    if (@args == 2) {
                        # replace "else{}" with "if true {}"
                        unless ($args[1]->test) {
                            $args[1]->test(sieve_node('test','true'));
                        }
                    }
                    $rule->delete_node();
                }
                else {
                    # no conditions left -- delete $arg from parents cond list
                    my @cmds = $parent->commands();
                    splice @cmds, $pos, 1;
                    $parent->commands(\@cmds);
                }
            };
            if ( $@ ) {
                $self->raise_error("Can't delete the rule: $@ $!");
            }
            else {
                $ruleset_id--;
                delete $self->{id};
                $self->save_script();
                $msg .= $self->{REQUEST}->query('cut') ? "cut." : "deleted.";
                $self->print_notification($msg);
            }
        }
        else {
            $self->rule($rule, $arg);
            $kw = "elsif";
        }
        if ( $self->{add_else} eq $rule->{id} ) {
            if ($arg->test) {
                # only we are inside if/elsif ( not else )
                my ($new_rule, $new_arg) = $self->add_condition($cmd, $i);
                $self->rule($new_rule, $new_arg);
                $kw = "elsif";
            }
        }
        elsif ( $self->{paste} eq $rule->{id} ) {
            my $clp = sieve_parse_file( $self->{clp_file} );
            $self->{id} = $ruleset_id;
            # if there is only one rule in the set ("if"-type) or if there are more rules but
            # the current one is "else" (the last one) then we paste new rule as stand-alone "if"
            if (scalar $cmd->args == 1 or not $arg->test) { 
                foreach ($clp->commands) {
                    my @args = $parent->commands();
                    $self->{new} = $_;
                    splice (@args, $pos+1, 0, $_);
                    $parent->commands(\@args);
                    $self->save_script();
                }
            }
            # otherwise we paste rules as alternatives
            else { 
                foreach ($clp->commands) {
                    foreach my $new_cmd ($_->args()) {
                        my ($new_rule, $new_arg) = $self->paste_as_alt($cmd, $i++, $new_cmd);
                        $self->rule($new_rule, $new_arg);
                    }
                }
            }
            $kw = "elsif";
        }
    }
}

sub add_if {
    my $self = shift;
    my $cmd = shift;
    my ($test, $cond, $if, @actions);
    $test = sieve_node( "test", "false" );
    push @actions, sieve_node( "action", "stop" );
    $cond = sieve_node( "cond", $test, "attr NAME=(New Rule)" );
    $cond->commands(@actions);
    $if = sieve_node( "if", "if", $cond );
    $self->{new} = $if;
    my @args = $cmd->commands();
    @args = add_cmd_before($if, ['stop'], @args);
    $cmd->commands(\@args);
    $self->save_script();
    return $if;
}

sub add_condition {
    my $self = shift;
    my $cmd = shift;
    my $i = shift;
    my ($test, $new, $new_rule, @actions);
    $test = sieve_node( "test", "false" );
    push @actions, sieve_node( "action", "stop" );
    $new = sieve_node( "cond", $test, "attr NAME=(New Rule)" );
    $new->commands(@actions);
    $self->{id} = $ruleset_id;
    my @args = $cmd->args();
    splice( @args, $i, 0 , $new );
    $cmd->args(@args);
    $new_rule = $self->{p}->rule($new,pos=>[$cmd,$i] );
    $self->save_rule($new_rule);
    $new_rule->{id} = $ruleset_id++;
    $new_rule->{new} = 1;
    $kw = "else";
    eval {
        $new_rule->save_node();
    };
    if ( $@ ) {
        $self->raise_error("Can't add the alternative: $@ $!");
    }
    else {
        $self->save_script();
    }
    return ($new_rule, $new);
}

sub paste_as_alt {
    my $self = shift;
    my $cmd = shift;
    my $i = shift;
    my $new_cmd = shift;
    my @args = $cmd->args();
    splice( @args, $i, 0, $new_cmd );
    $cmd->args(@args);
    my $new_rule = $self->{p}->rule($new_cmd,pos=>[$cmd,$i] );
    $self->save_rule($new_rule);
    $new_rule->{id} = $ruleset_id++;
    $kw = "else";
    eval {
        $new_rule->save_node();
    };
    if ( $@ ) {
        $self->raise_error("Can't paste the alternative rule: $@ $!");
    }
    else {
        $self->save_script();
    }
    return ($new_rule, $new_cmd);
}

sub rule_tb {
    my $self = shift;
    my $rule = shift;
    my @buttons = (
      {type=>'image', icon=>'save',  text=>'Save', name=>'save'},
      {type=>'image', icon=>'cancel',  text=>'Cancel', href=>"index.cgi"},
      {type=>'image', icon=>'copy', text=>'Copy', name=>'copy'},
      {type=>'image', icon=>'cut', text=>'Cut', name=>'cut'},
      {type=>'image', icon=>'delete', text=>'Delete', name=>'delete'},
      );
    push @buttons, {type=>'image', icon=>'paste', text=>'Paste', name=>'paste'}
        unless $self->empty_clipboard;
    push @buttons, {type=>'image', icon=>'add', text=>'Add Alternative', name=>'add_else'}
        unless $rule->{kw} eq "else";
    my $tb = PureMessage::Manager::Widget::ToolBar->new( buttons=> [\@buttons] );
    return $tb;
}

sub empty_clipboard {
    my $self = shift;
    return 1 unless -f $self->{clp_file};
    # if code snippet is not parsable -- do not display "paste" button
    eval {
        my $clp = sieve_parse_file( $self->{clp_file} );
    };
    return 0 unless $@;
}

sub get_tests_html {
    my $self = shift;
    my $rule = shift;
    my $html = "";
    my $i = 0;
    unless ( defined $self->{add_if} or defined $self->{add_else} ) {
        my @tests = $rule->tests();
        foreach ( @tests ) {
            $html .= '<tr>';
            if ($self->{save_failed} and $self->{new_test} == $_) {
                $html .= $self->get_test_html($_, "new", $#tests);
            }
            else {
                $html .= $self->get_test_html($_, $i++, $#tests);
            }
            $html .= '</tr>';
        }
    }
    if ( $self->{REQUEST}->query('add_test')) {
        $html .= '<tr>';
        $html .= $self->get_test_html(undef, "new");
        $html .= '</tr>';
    }
    if ( defined $self->{add_if} or defined $self->{add_else} ) {
        $html .= '<tr>';
        $html .= $self->get_test_html(undef, "0");
        $html .= '</tr>';
    }
    my $section_title;
    my $tb = button({type=>'image',  text=>'add test', name=>'add_test', icon=>'x',}, style=>'reverse');
    if ( $html ) {
        $section_title = "<tr><td class='tableRow' colspan=4><b>Tests:</b> $tb</td></tr>";
    }
    else {
        $section_title = "<tr><td class='tableRow' colspan=4><b>No tests in this rule, always match.</b> $tb</td></tr>";
    }
    return $section_title.$html;
}

sub get_test_html {
    my $self = shift;
    my $cmd = shift;
    my $pos = shift;
    my $can_move = shift;
    my $html = "";
    if ( my $test_html = $self->get_test_name_combo($cmd, $pos, $can_move) ) {
	my $hide_list;
	my $hide_arg;
	my $hide_opts;

        my $req_pop = ($cmd and $cmd->requires_popup());
	my $hide    = $req_pop || ($cmd and $cmd->arg_type(MIDDLE_COL) ne 'select');
        $html .= $test_html;
        $html .= $self->get_test_comp_combo($cmd, $pos, $hide);

        if ($req_pop) {
	    $hide_list = 1;
	    $hide_arg  = 1;
	    $hide_opts = "";
	}
	else {
	    $hide_list = ($cmd and $cmd->arg_selected) =~ /memberof/ ? 0 : 1;
	    $hide_arg = ($cmd and $cmd->arg_selected and $hide_list) ? 0 : 1;
	    $hide_opts = ($cmd and !$cmd->arg_type(MIDDLE_COL) and $cmd->args) ? "" : "display: none";
	}

	my $dis_arg = $hide_arg || !$hide_opts;
        my $value = "";
        if ($cmd) {
            $value = $hide_opts ? $cmd->arg_value(RIGHT_COL) : $cmd->arg_value(0);
        }
        $value = get_value($value);
        $html .= "<td valign=top class='tableRow' width='100%'><img src='/img/x.gif' width=1 height=1><span>"
              .  $self->get_arg_input($value, $pos, $hide_arg)
              .  $self->get_test_lists_combo($value, $pos, $hide_list)
              .  $self->get_test_args($cmd, $pos, $hide_opts)
              .  "</span></td>";
    }
    else {
        my $name = $cmd->name;
        $html .= <<EOT;
<td class='tableRow' width='100%' colspan=3>
<table width='100%' style="border-style: solid; border-width: 1px; border-color: #77756F"></tr><td class='tableRowAlt'>
The <b>$name</b> test contains complex arguments that cannot be displayed.
&nbsp;&nbsp;[<a href="/Policy/?source=1">view source</a>]
</td></tr></table>
</td>
EOT
    }
    return $html;
}

sub get_test_args {
    my $self = shift;
    my $cmd = shift;
    my $pos = shift;
    my $hide = shift;
    my $dis = $hide ? "disabled" : "";
    my $html = "";
    my $q_pos = $pos eq 'new' ? "'$pos'" : $pos;
    $html .= "<input class=button onClick=\"show_test_args($q_pos);\" type=button id=\"t_opt_$pos\" value='Arguments...' name=\"test_args_$pos\" style=\"$hide\">";
    $html .= "\n<input type=hidden id=\"t_ext_args_$pos\" name=\"t_ext_args_$pos\" $dis>";
    for (my $i=1; $i < $max_args; $i++) {
        my $value = "";
        if ( $cmd and $i < $cmd->args() ) {
            $value = $cmd->arg_value($i);
            $value = join(" ", @$value) if ref($value) eq "ARRAY";
        }
        $html .= "<input type=hidden id=\"arg$i\_$pos\" name=\"test_arg$i\_$pos\" value=\"".encode_entities($value,'&<>"\/\'')."\">\n";
    }
    return $html;
}

sub get_arg_input {
    my $self = shift;
    my $value = shift;
    my $pos = shift;
    my $hide = shift;
    $hide = "display: none" if $hide;
    my $dis = shift;
    $dis = "disabled" if $dis;
    my $html = "";
    $html .= "<input id='arg0_$pos' $dis "
          .  " name=\"test_arg0_$pos\" value=\""
          .  html_escape($value)
          .   "\""
          .  " style=\"width: 100%; $hide\">";
    return $html;
}

sub get_test_lists_combo {
    my $self = shift;
    my $value = encode_entities(shift,'&<>"\/\'');
    my $pos = shift;
    my $hide = shift;
    $hide = "display: none" if $hide;
    my $dis = $hide ? "disabled" : "";
    my $html = "";
    my $found;
    $html .= "<select $dis id='list_$pos' onChange=\"update_test_row('$pos')\" "
          .  " name=\"test_list_$pos\""
          .  " style=\"$hide\">";
    use PureMessage::List;
    use PureMessage::MultiList;    
    for (sort {$a->{name} cmp $b->{name}} (
                                        PureMessage::List->Available,
                                        PureMessage::MultiList->Available)) {
        $html .= "<option value=\"" . $_->{id} . "\" ";
        if ( $value and $value eq $_->{id} ) {
            $html .= "selected";
            $found++;
        }
        $html .= " >";
        $html .= $_->{name};
    }
    unless ( $found ) {
        if ( $value ) {
            $html .= "<option value='$value' selected>".$value;
        }
        else {
            $html .= "<option value='' selected>&lt;select the list&gt;";
        }
    }
    $html .= "</select>";
    return $html;
}

sub get_test_name_combo {
    my $self = shift;
    my $cmd = shift;
    my $pos = shift;
    my $can_move = shift;
    my $html = "";
    my $found;
    my $hide = shift;
    $hide = $hide ? "display: none;" : "";
    my $dis = $hide ? "disabled" : "";
    $html .= "<td valign=top class='tableRow' nowrap=\"nowrap\">"
          .  "<select id='name_$pos' onChange=\"update_test_row('$pos')\" "
          .  "name=\"test_name_$pos\">";
    foreach ( sort { $a->label cmp $b->label } $self->{p}->test_list ) {
        $html .= "<option  value=\"" . $_->name . "\" ";
        if ( $cmd and $_->name eq $cmd->name ) {
            $html .= "selected";
            $found++;
        }
        $html .= " >";
        $html .= $_->label;
    }
    unless ( $found ) {
        if ( $cmd ) {
            return "";
        }
        else {
            $html .= "<option value='' selected>&lt;select the test&gt;";
        }
    }
    $html .= "</select>";
    my $at = ref($cmd) ? $cmd->name : "";
    $html .= "<br><a id='tests_help_$pos' href='$tests_help#$at'  title='Help'"
          .  "onClick=\"window.open(this.href, 'docs', 'width=550, height=400, toolbar=no, resizable=yes, scrollbars=yes'); return false;\">"
          .  "<img border=0 src='/img/q_mark.gif'></a>";
    if ( $can_move ){
        $html .= "<a href='?pos=$pos&move=test&id=$self->{id}' title='Move up'><img border=0 src='/img/r_up.gif'></a>";
        $html .= "<a href='?pos=$pos&move=test&d=down&id=$self->{id}' title='Move down'><img border=0 src='/img/r_down.gif'></a>";
    }
    $html .= "<a href='?del_test=$pos&id=$self->{id}' title='Delete this test'><img border=0 src='/img/row_x.gif'></a>" unless $pos eq "new";
    $html .= "</td>";
    return $html;
}

sub get_test_comp_combo {
    my $self = shift;
    my $cmd = shift;
    my $pos = shift;
    my $hide = shift;
    $hide = "display: none" if $hide;
    my $html = "";
    # print a select input
    $html .= "\n<td valign=top class='tableRow'><img src='/img/x.gif' width=1 height=1>"
          .  "<select id='comp_$pos' onChange=\"update_test_row('$pos')\" "
          .  " name=\"test_op_$pos\""
          .  " style=\"$hide\">";
    if ( $cmd and my $options = $cmd->arg_options() ) {
        foreach my $key (sort keys %$options) {
            $html .= "\n\t<option value=\"$key\" ";
            $html .= "selected" if $key eq $cmd->arg_selected;
            $html .= " >";
            $html .= $options->{$key};
        }
        $html .= "\n</select></td>\n";
    }
    return $html;
}

sub get_anyall_html {
    my $self = shift;
    my $rule = shift;
    my $items = {   
                    0=>'If ALL criteria are met',
                    1=>'If ANY criteria are met',
                    2=>'If NO criteria are met',
                    3=>'Not ALL criteria are met',
                };
    my $html = "";
    $html .= "<select name=\"anyall\" class=\"updateAvailable\">";
    foreach ( keys %$items ) {
        $html .= "<option name='anyall' value='$_' ";
        $html .= "selected" if $_ == $rule->anyof;
        $html .= " >$items->{$_}</option>";
    }
    $html .= "</select>";
    return $html;
}

sub get_actions_html {
    my $self = shift;
    my $rule = shift;
    my $pos = shift;
    my $html = "";
    my $i = 0;
    my @actions = $rule->actions();
    foreach ( @actions ) {
        $html .= '<tr>';
        if ($self->{save_failed} and $self->{new_action} == $_) {
            $html .= $self->get_action_html($_, "", $#actions);
        }
        else {
            $html .= $self->get_action_html($_, $i, $#actions);
        }
        $html .= '</tr>';
        $i++;
    }
    if ( $self->{REQUEST}->query('add_action') ) {
        $html .= '<tr>';
        $html .= $self->get_action_html();
        $html .= '</tr>';
    }
    my $tba = button({type=>'image', text=>'add action', name=>'add_action', icon=>'x',}, style=>'reverse');
    my $tbr = button({type=>'image', text=>'add rule', name=>'add_if', icon=>'x',}, style=>'reverse');
    my $title = $html ? "Execute actions and rules:" : "No actions defined.";
    $html = "<tr><td class='tableRow' colspan=4><b>$title</b> $tba $tbr</td></tr>" . $html;
    return $html;
}

sub get_action_html {
    my $self = shift;
    my $cmd = shift;
    my $pos = shift;
    $pos = "" unless defined $pos;
    my $can_move = shift;
    my $html = "";
    $html .= $self->get_action_name_combo($cmd, $pos, $can_move);
    $html .= "<td valign=top class='tableRow' colspan=2><table width='100%' cellspacing=0 cellpadding=0><tr>";
    my $args_count = $cmd ? scalar $cmd->args : 0;
    $html .= "<td class='tableRow' width='1'>"
          .  $self->get_action_args($cmd, $pos, ( $args_count > 2 ? "" : "display: none"))
          .  "</td>";
    $html .= "<td class='tableRow' width='40%'>"
          .  $self->get_action_arg($cmd, $pos, 0, 
                    (($args_count > 0 and $args_count <= 2) ? "" : "display: none"))
          .  "</td><td class='tableRow' width='1'>&nbsp;</td>";
    $html .= "<td class='tableRow' width='60%'>"
          .  $self->get_action_arg($cmd, $pos, 1, ($args_count == 2 ? "" : "display: none"))
          .  "</td>";
    $html .=  "</tr></table></td>";
    return $html;
}

sub get_value {
    my $value = shift;
    if (ref($value) eq "ARRAY") {
        if (@$value > 1) {
            $value = '"' . join ('", "', @{$value}) . '"';
        }
        else {
            $value = $value->[0];
        }
    }
    return $value;
}

sub get_action_arg {
    my $self = shift;
    my $cmd = shift;
    my $pos = shift;
    my $i = shift;
    my $hide = shift;
    my $dis = $hide ? "disabled" : "";
    my $value = ($cmd and scalar $cmd->args > $i) ? $cmd->arg_value($i) : "";
    $value = get_value($value);
    my $html = "";
    $html .=  "<input id=\"a_arg$i\_$pos\" name=\"action_arg$i\_$pos\" value=\"".html_escape($value)."\" style=\"width: 100%; $hide\">";
    return $html;
}

sub get_action_args {
    my $self = shift;
    my $cmd = shift;
    my $pos = shift;
    my $hide = shift;
    my $dis = $hide ? "disabled" : "";
    my $html = "";
    $html .= "<input class=button onClick=\"show_action_args($pos);\" type=button id=\"a_opt_$pos\" value='Arguments...' name=\"action_args_$pos\" style=\"$hide\">";
    for (my $i=2; $i < $max_args; $i++) {
        my $value = "";
        if ( $cmd and $i < $cmd->args() ) {
            $value = $cmd->arg_value($i);
            $value = join(" ", @$value) if ref($value) eq "ARRAY";
        }
        $html .= "<input type=hidden id=\"a_arg$i\_$pos\" name=\"action_arg$i\_$pos\" value=\"$value\">\n";
    }
    return $html;
}

sub get_action_name_combo {
    my $self = shift;
    my $cmd = shift;
    my $pos = shift;
    my $can_move = shift;
    my $found = 0;
    my $html = "";
    $html .= "\n<td valign=top class='tableRow'><select id='a_name_$pos'"
           . " onChange='update_action_row($pos);' "
           . " name=\"action_name_$pos\" style=\"width: 100%\">";
    foreach ( sort {$a->label cmp $b->label} values %{$self->{_actions}} ) {
        $html .= "\n\t<option value=\"".$_->name.'"';
        if ( $cmd and $_->name eq $cmd->name ) {
            $found = 1;
            $html .= " selected " ;
        }
        $html .= " >";
        $html .= $_->label;
    }
    unless ( $found ) {
        if ( $cmd ) {
            $html .= "<option value=\"".$cmd->name()."\" selected>".$cmd->name();
        }
        else {
            $html .= "<option value='' selected>&lt;select the action&gt;";
        }
    }
    $html .= "\n</select><br>\n";
    my $at = ref($cmd) ? $cmd->name : ""; 
    $html .= "<a id='actions_help_$pos' href='$actions_help#$at' "
          .  "onClick=\"window.open(this.href, 'docs', 'width=550, height=400, toolbar=no, resizable=yes, scrollbars=yes'); return false;\">"
          .  "<img border=0 src='/img/q_mark.gif'></a>";
    if ( $can_move ){
        $html .= "<a href='?pos=$pos&move=action&id=$self->{id}' title='Move up'><img border=0 src='/img/r_up.gif'></a>";
        $html .= "<a href='?pos=$pos&move=action&d=down&id=$self->{id}' title='Move down'><img border=0 src='/img/r_down.gif'></a>";
    }
    $html .= "<a href='?del_action=$pos&id=$self->{id}'><img border=0 src='/img/row_x.gif'></a>" unless $pos eq "";
    $html .= "</td>";
    return $html;
}

sub delete_test {
    my $self = shift;
    my $rule = shift;
    my $test = shift;
    my @tests = $rule->tests;
    if ( $rule->{kw} eq "if" and scalar @tests < 2 ) {
        $self->raise_error("Can't delete: there should be at least one test in the top rule.");
        return;
    }
    elsif ( $rule->{pos}[1]+1 < scalar $rule->{pos}[0]->args 
            and scalar @tests < 2 ) {
        $self->raise_error("Can't delete: there should be at least one test in this alternative rule.");
        return;
    }
    splice( @tests, $test, 1 );
    $rule->{tests} = \@tests;
    $self->save_rule($rule);
}

sub delete_action {
    my $self = shift;
    my $rule = shift;
    my $action = shift;
    my @actions = $rule->actions;
    splice( @actions, $action, 1 );
    $rule->{actions} = \@actions;
    $self->save_rule($rule);
}

sub reorder {
    my $self = shift;
    my $rule = shift;
    my $what = $self->{REQUEST}->query('move');
    my $pos = $self->{REQUEST}->query('pos');
    my $where = $self->{REQUEST}->query('d') eq "down" ? 1 : -1;
    return unless $what and $where and defined $pos;
    $what = $what eq "action" ? "actions" : "tests";
    my @elems = $rule->$what; # list of actions or tests of the rule
    my $elem = $elems[$pos];
    splice( @elems, $pos, 1 ); # remove selected action/test first
    my $new = $pos + $where;
    $new = 0 if $new > $#elems + 1;
    # pust selected action/test back in new position
    if ( $new < 0 ) {
        push @elems, $elem
    }
    else {
        splice( @elems, $new, 0, $elem );
    }
    $rule->{$what} = \@elems;
    $self->save_rule($rule);
}

sub edit_rule {
    my $self = shift;
    my $rule = shift;
    my $html = "";
    $html .= "<input type=hidden name=modified value=1>";
    $html .= "<input type=hidden name=id value=".$self->{id}.">";
    unless ( $rule->{new} 
            or defined $self->{add_if} 
            or defined $self->{paste} ) {
        $self->rewrite_rule($rule);
        # XXX ERROR CHECK HERE
        $self->delete_test($rule, $self->{REQUEST}->query('del_test'))
            if defined $self->{REQUEST}->query('del_test');
        $self->delete_action($rule, $self->{REQUEST}->query('del_action'))
            if defined $self->{REQUEST}->query('del_action');
        $self->reorder($rule);
        $self->save_rule($rule) if $self->{save};
    }
    $html .= "<table width='100%' cellspacing=0 cellpadding=4 border=0>";
    $html .= "<tr><td colspan=4 class='tableRow'><input name=name value='".html_escape($rule->name)."' class='stretch'></td></tr>";
    $html .= $self->get_tests_html($rule);
    if ( $rule->tests > 1 
        or ($rule->tests > 0 and $self->{REQUEST}->query('add_test')) ) {
        $html .= '<tr><td colspan=4 class="tableRow">'.$self->get_anyall_html($rule).'</td></tr>';
    }
    $html .= $self->get_actions_html($rule);
    $html .= "</table>";
    return $html;
    
}

sub is_array {
    my $value = shift;
    # strip leading and trailing spaces
    $value =~ s/^\s+//g;
    $value =~ s/\s+$//g;

    # compile list from comma-separated values
    my @tokens = split(/\s*,\s*/, $value);
    return $value unless @tokens; # the value could've been an empty string
    foreach (@tokens) {
        # list elements must be doublequoted, otherwise it's a string
        return $value unless s/^\"// and s/\"$//;
    }
    return \@tokens;
}

sub rewrite_test {
    my ($self, $t, $i) = @_;
    if ( my $new_op = $self->{REQUEST}->query("test_op_$i") ) {
        $t->arg_selected($new_op);
    }
    if ( $self->{REQUEST}->query("test_op_$i") =~ /memberof/ ) {
        if ( my $new_val = $self->{REQUEST}->query("test_list_$i") ) {
            $t->arg_value(RIGHT_COL, $new_val);
        }
    }
    if (defined($self->{REQUEST}->query("t_ext_args_$i"))) {
        for ( my $p=0; $p < $t->args; $p++ ) {
            my $new_val = $self->{REQUEST}->query("test_arg$p\_$i");
            if ( defined $new_val ) {
                if ($t->arg_type($p) =~ /^checkbox_(.*)$/ and not length($new_val) ) {
                    $new_val = undef;
                }
                $new_val = is_array($new_val);
                $t->arg_value($p, $new_val);
            }
        }
    }
    else {
        my $new_val = $self->{REQUEST}->query("test_arg0\_$i");
        if ( defined $new_val ) {
            $new_val = is_array($new_val);
            $t->arg_value(RIGHT_COL, $new_val);
        }
    }
}

sub rewrite_rule {
    my $self = shift;
    my $rule = shift;
    my $i = 0;
    # update rule name
    if ( my $new_name = $self->{REQUEST}->query("name") ) {
        $rule->name($new_name);
    }
    my @test = $rule->tests();
    foreach my $t ( $rule->tests() ) {
        if ( my $new_name = $self->{REQUEST}->query("test_name_$i") ) {
            $t->name($new_name);
        }
        $self->rewrite_test($t, $i);
        $i++;
    }
    # add new test
    if ( my $new_name = $self->{REQUEST}->query("test_name_new") ) {
        my $t = $rule->add_test($new_name);
        $self->{new_test} = $t;
        $self->rewrite_test($t, 'new');
    }
    # update anyall
    if ( defined $self->{REQUEST}->query("anyall") ) {
        $rule->anyof($self->{REQUEST}->query("anyall"));
    }
    # update actions
    $i = 0;
    foreach my $a ( $rule->actions() ) {
        if ( my $new_name = $self->{REQUEST}->query("action_name_$i") ) {
            $a->name($new_name);
        }
        my $cmd = $self->{_actions}{$a->name};
        if ($cmd) {
            for ( my $p=0; $p < $cmd->args; $p++ ) {
                my $new_val = $self->{REQUEST}->query("action_arg$p\_$i");
                if ( defined $new_val ) {
                    if ($a->arg_type($p) =~ /^checkbox_(.*)$/ and not length($new_val) ) {
                        $new_val = undef;
                    }
                    $new_val = is_array($new_val);
                    $a->arg_value($p, $new_val);
                }
            }
        }
        $i++;
    }
    # add new actions
    if ( my $new_name = $self->{REQUEST}->query("action_name_") ) {
        my $cmd = $rule->add_action($new_name);
        $self->{new_action} = $cmd;
        if ($cmd) {
            for ( my $p=0; $p < $cmd->args; $p++ ) {
                my $new_val = $self->{REQUEST}->query("action_arg$p\_");
                if ( defined $new_val ) {
                    if ($cmd->arg_type($p) =~ /^checkbox_(.*)$/ and not length($new_val) ) {
                        $new_val = undef;
                    }
                    $new_val = is_array($new_val);
                    $cmd->arg_value($p, $new_val);
                }
            }
        }
    }
    
}

sub save_rule {
    my $self = shift;
    my $rule = shift;
    eval {
        $rule->save_node();
    };
    if ( $@ ) {
        $self->{save_failed} = 1;
        $self->raise_error("Can't save the rule: $@ $!");
    }
    else {
        $self->save_script();
    }
}

sub save_script {
    my $self = shift;
    # check if source was modified
    open ( my $O, "<$self->{file}");
    my $original = join("", <$O>);
    close $O;
    return if $original eq sieve_text($self->{script});
    $self->{file} .= ".edit" unless $self->{file} =~ /\.edit/;
    open ( my $S, ">$self->{file}" ) || die "Can't open file for writing $self->{file}.edit";
    sieve_print( $self->{script}, $S );
    $self->commit_dlg();
}

sub javascript {
    my $self = shift;
    my $html;
    # prepare comparator arrays
    $html .= "\tvar t_ops = new Array();\n";
    foreach my $t ( $self->{p}->test_list() ) {
        $html .= "\tt_ops['".$t->name."'] = new Array ();\n";
        if (my $options = $t->arg_options()) {
            my $i = 0;
            foreach my $k (sort keys %$options) {
                $html .= "\tt_ops['".$t->name."'][$i] = '$k';\n"; $i++;
                $html .= "\tt_ops['".$t->name."'][$i] = '$options->{$k}';\n"; $i++;
            }
        }
    }
    $html .= "\tvar t_args = new Array();\n";
    foreach my $t ( $self->{p}->test_list() ) {
	my $type;
	if ($t->args) {
	    my $col = $t->arg_type(MIDDLE_COL);
	    if (!$col || $t->requires_popup) {
		$type = 'popup';
	    }
	    else {
		$type = $col;
	    }
	}
	else {
	    $type = '';
	}
        $html .= "\tt_args['".$t->name."'] = '$type';\n";
    }
    $html .= "\tvar a_args = new Array();\n";
    foreach my $a ( values %{$self->{_actions}} ) {
        $html .= "\ta_args['".$a->name."'] = ".$a->args().";\n";
    }
    # remember argument names for actions with 1 or 2 arguments
    $html .= "\tvar a_arg1 = new Array();\n";
    $html .= "\tvar a_arg2 = new Array();\n";
    foreach my $a ( values %{$self->{_actions}} ) {
        $html .= "\ta_arg1['".$a->name."'] = '" . $a->arg_label(0) . "';\n" 
            if $a->args() > 0 and $a->args() < 3;
        $html .= "\ta_arg2['".$a->name."'] = '" . $a->arg_label(1) . "';\n" 
            if $a->args() > 1 and $a->args() < 3;
    }
    $html .= <<EOT 
    function update_test_row(id)
    {
        test = document.getElementById('name_'+id);
        ops = document.getElementById('comp_'+id);
        opt = document.getElementById('t_opt_'+id);
        arg1 = document.getElementById('arg0_'+id);
        arg2 = document.getElementById('list_'+id);
        ext =  document.getElementById('t_ext_args_'+id);
        
        var s = ops.selectedIndex;
        ops.options.length = 0;
        var t = test.options[test.selectedIndex].value;
        
        if (tests_help = document.getElementById('tests_help_'+id)) {
            tests_help.href = '$tests_help' + '#' + t;
        }
        
        if (t) {
            for ( i=0; i < t_ops[t].length; i+=2 ) 
            {
                ops.options[i/2] = new Option(t_ops[t][i+1],t_ops[t][i]);
            }
            if ( s > 0 ) {
                ops.options.selectedIndex = s;
            }
        }
        if ( arg1.value ) {
            arg1.value = "";
        }
        if ( t_args[t] ) {
            if ( t_args[t] == "select") {
                opt.disabled = 1;
                ext.disabled = 1;
                opt.style.display = "none";
                ops.disabled = 0;
                ops.style.display = "block";
                if ( ops.value.substring(0, 8)  == 'memberof' ) {
                    arg2.disabled = 0;
                    arg2.style.display = "block";
                    arg1.disabled = 1;
                    arg1.style.display = "none";
                }
                else {
                    arg1.disabled = 0;
                    arg1.style.display = "block";
                    arg2.disabled = 1;
                    arg2.style.display = "none";
                }
            }
            else {
                if (  t_args[t] == "popup" ) {
                    opt.disabled = 0;
                    ext.disabled = 0;
                    opt.style.display = "block";
                    arg1.style.display = "none";
                }
                else {
                    opt.disabled = 1;
                    ext.disabled = 1;
                    opt.style.display = "none";
                    arg1.style.display = "block";
                }
                arg1.disabled = 0;
                arg2.disabled = 1;
                arg2.style.display = "none";
                ops.disabled = 1;
                ops.style.display = "none";
            }
        }
        else {
            ops.disabled = 1;
            ops.style.display = "none";
            arg1.disabled = 1;
            arg1.style.display = "none";
            arg2.disabled = 1;
            arg2.style.display = "none";
            opt.disabled = 1;
            opt.style.display = "none";
            ext.disabled = 1;
        }
    }

    function update_action_row(id)
    {
        if (id == null) { id = '' }
        action = document.getElementById('a_name_'+id);
        arg1 = document.getElementById('a_arg0_'+id);
        arg2 = document.getElementById('a_arg1_'+id);
        opt = document.getElementById('a_opt_'+id);
        var i = 0;
        
        var t = action.options[action.selectedIndex].value;
        
        if (actions_help = document.getElementById('actions_help_'+id)) {
            actions_help.href = '$actions_help' + '#' + t;
        }
        
        for ( i=0; i < $max_args; i++) {
            arg = document.getElementById('a_arg'+i+'_'+id);
            if ( arg ) {
                if ( i == 0 && a_arg1[t] ) {
                    arg.value = '(' + a_arg1[t] + ')';
                }
                else if ( i == 1 && a_arg2[t] ) {
                    arg.value = '(' + a_arg2[t] + ')';
                }
                else {
                    arg.value = "";
                }
            }
        }
        
        if ( a_args[t] > 0 ) {
            if ( a_args[t] > 2 ) {
                opt.disabled = 0;
                opt.style.display = "block";
                arg1.disabled = 0;
                arg1.style.display = "none";
                arg2.disabled = 0;
                arg2.style.display = "none";
            }
            else if ( a_args[t] > 1 ) {
                opt.disabled = 1;
                opt.style.display = "none";
                arg1.disabled = 0;
                arg1.style.display = "block";
                arg2.disabled = 0;
                arg2.style.display = "block";
            }
            else {
                opt.disabled = 1;
                opt.style.display = "none";
                arg1.disabled = 0;
                arg1.style.display = "block";
                arg2.disabled = 1;
                arg2.style.display = "none";
            }
        }
        else {
            arg1.disabled = 1;
            arg1.style.display = "none";
            arg2.disabled = 1;
            arg2.style.display = "none";
            opt.disabled = 1;
            opt.style.display = "none";
        }
    }

    function show_action_args(id)
    {
        if (id == null) { id = '' }
        var a = document.getElementById('a_name_'+id);
        var action = a.options[a.selectedIndex].value;
        window.open('args.cgi?id='+id+'&action='+action, 'args', 'width=400, height=500, toolbar=no');
    }

    function show_test_args(id)
    {
        if (id == null) { id = '' }
        var t = document.getElementById('name_'+id);
        var test = t.options[t.selectedIndex].value;
        window.open('args.cgi?id='+id+'&test='+test, 'args', 'width=400, height=500, toolbar=no');
    }
    
    function custom_onSubmit (f)
    {
        if (f && f.save != null && f.save.value == "save") {
            for(var e = 0; e < f.elements.length; e++) {
                var element = f.elements[e];
                if ( element != null && element.name.substr(0, 12) == "action_name_") {
                    var id = element.name.substr(12, element.name.length);
                    var a = document.getElementById('a_arg0_'+id);
                    if (a && a_arg1[element.value] && (a.value == '(' + a_arg1[element.value] + ')')) {
                        alert("Please enter a value for the '" + a_arg1[element.value] + "' argument.  Click the '?' icon below the associated action for help.");
                        return false;
                    }
                    a = document.getElementById('a_arg1_'+id);
                    if (a && a_arg2[element.value] && (a.value == '(' + a_arg2[element.value] + ')')) {
                        alert("Please enter a value for the '" + a_arg2[element.value] + "' argument.  Click the '?' icon below the associated action for help.");
                        return false;
                    }
                }
            }
        }
        return true;
    }
EOT
}

1;
