package PureMessage::Policy::Base;
use base 'PureMessage::Policy::Plugin';
use strict;
use Carp qw(croak);
use PureMessage::Policy::Events qw(event_valid);
use PureMessage::Config qw(%config);
use Mail::Address;
use PureMessage::MessageInfo;
use PureMessage::Queue 'queue_mail';
use PureMessage::Policy::Plugin::Utils qw(get_recipients get_matcher 
					  recode_header normalize_addrs get_action_data);
use PureMessage::UTF8::Config qw(%utf8_config);
use ActiveState::I18N::UTF8 qw(utf8_header encode_utf8_header);
use ActiveState::UTF8 qw(is_valid_utf8);
use PureMessage::Marks qw(is_external_mark);

my %commands = (
    header => \&header,
    address => \&address,
    size => \&size,
    reject => \&reject,
    keep => \&keep,
    exists => \&exists,
    envelope => \&envelope,
    pmx_header_size => \&pmx_header_size,
    pmx_body_size => \&pmx_body_size,
    pmx_quarantine => \&pmx_quarantine,
    pmx_relay => \&pmx_relay,
    redirect => \&redirect,
    pmx_mark => \&pmx_mark,
    pmx_mark1 => \&pmx_mark1,
    pmx_test_mark => \&pmx_test_mark,
    discard => \&discard,
    pmx_tempfail => \&pmx_tempfail,
    pmx_file => \&pmx_file,
    pmx_log => \&pmx_log,
    pmx_add_recipient => \&pmx_add_recipient,
    # XXX: add in to enable changing envelope from 
    # pmx_change_from => \&pmx_change_from,
    pmx_add_header => \&pmx_add_header,
    pmx_delete_header => \&pmx_delete_header,
    pmx_replace_header => \&pmx_replace_header,
    pmx_map_recipients => \&pmx_map_recipients,
    pmx_number_of_recipients => \&pmx_number_of_recipients,
    pmx_forward => \&pmx_forward,
    pmx_notify => \&pmx_notify,
    pmx_deliver_for => \&pmx_deliver_for,
    pmx_mta_macro => \&pmx_mta_macro,
    true => \&true,
    false => \&false,
               );
my %templates = (
    SUBJECT => \&SUBJECT,
    MESSAGE_SIZE => \&MESSAGE_SIZE,
    BODY_SIZE => \&BODY_SIZE,
    HEADER_SIZE => \&HEADER_SIZE,
    QUEUE_ID => \&QUEUE_ID,
    SENDER_IP => \&SENDER_IP,
    DATETIME => \&DATETIME,
    DATETIME_GMT => \&DATETIME_GMT,
    ENVELOPE_FROM => \&ENVELOPE_FROM,
    ENVELOPE_TO => \&ENVELOPE_TO,
    HEADER_FROM => \&HEADER_FROM,
    HEADER_TO => \&HEADER_TO,
    HEADER_CC => \&HEADER_CC,
    HEADER_DATE => \&HEADER_DATE,
               );

sub new {
    my $class = shift;
    my $dict = shift;
    my %opts = @_;
    return bless \%opts, $class;
}

sub commands {
    return (%commands, %templates);
}

sub templates {
    return %templates;
}

sub _str_compare {
    my $msg = shift;
    my $args = shift;
    my @strs = @_;
    my $matcher;
    if (defined $msg) {
        $matcher = get_matcher($msg, $args);
    }
    else {
        $matcher = $args->matcher();
    }
    croak "no match function found!" unless $matcher;
    return $matcher->(@strs);    
}

sub _num_compare {
    my $num = shift;
    my $args = shift;
    my $i = shift;
    $args->tag('over') and return ($num > $args->param($i));
    $args->tag('under') and return ($num < $args->param($i));
    return ($num == $args->param($i));
}

sub true {
    return 1; 
}

sub false {
    return 0;
}

sub header {
    my $msg = shift;
    my $args = shift;
    
    for (@{$args->param(0)}) {
        next unless $msg->header_exists($_);
        return 1 if _str_compare($msg, $args, $msg->header_decoded($_));
    }
    return 0;
}

sub address {
    my $msg = shift;
    my $args = shift;

    local $Mail::Address::VERBOSE = 0;
    for (@{$args->param(0)}) {
        next unless $msg->header_exists($_);
        for ($msg->header_decoded($_)) {            
            my @addrs;
            if ($args->tag('localpart')) {
                @addrs = map $_->user, Mail::Address->parse($_);
            }
            elsif ($args->tag('domain')) {
                @addrs = map $_->host, Mail::Address->parse($_);
            }
            else {
                @addrs = map $_->address, Mail::Address->parse($_);
            }
            
            # Handle sub-addressing
            unless ($args->tag('detail')) {
                @addrs = normalize_addrs(@addrs);  
            }

            return 1 if _str_compare($msg, $args, @addrs);
        }
    }
    return 0;
}

sub size {
    my $msg = shift;
    my $args = shift;
    return _num_compare($msg->message_size,$args,0);
}

sub reject {
    my $msg = shift;
    my $args = shift;
    $msg->delivery_action('reject',
                          $args->param(0)->[0],
                          $args->tag('rcode') ? $args->tag('rcode')->{params}[0]{value} : 550,
                          $args->tag('xcode') ? $args->tag('xcode')->{params}[0]{value}[0] : '');
}

sub discard {
    my $msg = shift;
    my $args = shift;
    $msg->delivery_action('discard');   
}

sub pmx_tempfail {
    my $msg = shift;
    my $args = shift;
    $msg->delivery_action('tempfail');   
}

sub keep {
    my $msg = shift;
    my $args = shift;
    $msg->delivery_action('keep');
}

sub exists {
    my $msg = shift;
    my $args = shift;
    foreach (@{$args->param(0)}) {
        return 1 if $msg->header_exists($_);
    }
    return 0;
}

sub envelope {
    my $msg = shift;
    my $args = shift;
    
    $msg->assert_event('envelope');
    
    local $Mail::Address::VERBOSE = 0;
    foreach (@{$args->param(0)}) {
        my ($lc_key, @raw_addrs) = lc $_;
        if ($lc_key eq 'to'){
            @raw_addrs = get_recipients($msg);
        }
        elsif ($lc_key eq 'from') {
            @raw_addrs = $msg->minfo_value("Envelope-From");
        }
        else {
            next;
        }

        my @addrs;
        if ($args->tag('localpart')) {
            @addrs = map $_->user, Mail::Address->parse(@raw_addrs);
        }
        elsif ($args->tag('domain')) {
            @addrs = map $_->host, Mail::Address->parse(@raw_addrs);
        }
        else {
            @addrs = map $_->address, Mail::Address->parse(@raw_addrs);
        }
         
        # Handle sub-addressing
        unless ($args->tag('detail')) {
            @addrs = normalize_addrs(@addrs);  
        }
       
        return 1 if _str_compare($msg, $args, @addrs);
    }

    if($args->param(0)->[0] eq 'group') {
        my @groups = ();
        my $direction = $msg->minfo_value('Direction');
        my @_groups = ($direction eq 'Outbound')
            ? $msg->minfo_value('Sender-Group')
            : $msg->minfo_value('Recipient-Group');
        if(@_groups) {
            my %_groups = ();
            for my $_group (@_groups) {
                my ($recipient, $group) = split /\|/, $_group;
                $_groups{$recipient} = $group;
            }
            if($direction eq 'Outbound') {
                push @groups, (values %_groups)[0];
            }
            else {
                for my $recipient (get_recipients($msg)) {
                    push @groups, $_groups{$recipient} if $_groups{$recipient};
                }
            }
        }
        return 1 if _str_compare($msg, $args, @groups);
    }

    return 0;
}

sub pmx_number_of_recipients {
    my $msg = shift;
    my $args = shift;
    $msg->assert_event('envelope');
    my $count = $msg->get_attr('envrcpt_count') || 1;
    $msg->log('DEBUG',"pmx_number_of_recipients: $count recipient(s)");    
    return _num_compare($count,$args,0);
}

sub pmx_header_size {
    my $msg = shift;
    my $args = shift;
    return _num_compare($msg->header_size,$args,0);
}

sub pmx_body_size {
    my $msg = shift;
    my $args = shift;
    return _num_compare($msg->body_raw_size,$args,0);
}

sub pmx_quarantine {
    my $msg = shift;
    my $args = shift;
    $msg->assert_event('eom');
    $msg->delivery_action('quarantine', $args->param(0)->[0]);
}

sub pmx_file {
    my $msg = shift;
    my $args = shift;
    $msg->file($args->param(0)->[0]);
}

sub pmx_forward {
    my $msg = shift;
    my $args = shift;
    
    $msg->assert_event('eom');
    my $to = $args->param(0);
    queue_mail(tempfile => $msg->implode_to_tempfile(), to=>$to, minfo => $msg->minfo());
    $msg->log('DEBUG','pmx_forward: Queued message for ' . join (', ', @$to));
}

sub pmx_notify {
    my $msg = shift;
    my $args = shift;

    $msg->assert_event('eom');

    my @orig;
    if ($args->tag('sender')) {
        @orig = $msg->minfo_value("Envelope-From");
    }
    else {
        @orig = $msg->minfo_value("Envelope-To");
    }

    my @recips;
    my $transformer = $args->transformer();
    if ($transformer) {
        @recips = $transformer->(@orig);
    }
    else {
        @recips = @orig;
    }

    my $data = get_action_data($msg, $args);
    return if (!$data);

    my $admin_addr;
    if ($args->tag('groupfile')) {
        my $notify_msg = PureMessage::Policy::Message->new(quarantine => $msg->{base}{quarantine});
        $notify_msg->print($data);
        $notify_msg->event('eom');
        $admin_addr = $notify_msg->header('From');
    }

    my $tmp = $msg->new_tempfile;

    # add the UTF8 content type if the message contains non-ascii characters
    if ($data =~ /[\x80-\xff]/) {
        my $notify_msg = PureMessage::Policy::Message->new(quarantine =>
                                                      $msg->{base}{quarantine});
        $notify_msg->print($data);
        $notify_msg->event('eom');
        
        unless ($notify_msg->header_exists('Content-Type')) {
            # add required content type headers
            $notify_msg->header_add('Content-Type', "text/plain; charset=UTF-8");
            $notify_msg->header_add('Content-Transfer-Encoding', '8bit');
        }
        
        # callback for imploade
        my $cb = sub {
            my $data = shift;
            $tmp->print($data);
        };
        
        # save the message to the temp file.
        $notify_msg->implode($cb);
    }
    else {
        $tmp->print($data);
    }
    
    local $Mail::Address::VERBOSE = 0;
    $admin_addr = (Mail::Address->parse($config{admin}))[0]->address if (!$admin_addr);
    queue_mail(tempfile=>$tmp, to=>\@recips, from=>$admin_addr);    
    $msg->log('DEBUG','pmx_notify: Queued message for ' . join (', ', @recips));
}

sub pmx_map_recipients {
    my $msg = shift;
    my $args = shift;
    
    $msg->assert_event('envelope');
    my @orig = $msg->minfo_value("Envelope-To");
    my $transformer = $args->transformer or croak 'no transformer function found';
    my @new = $transformer->(@orig);
    $msg->minfo_value("Envelope-To", @new);
    $msg->log('DEBUG','pmx_map_recipients: (' . join(', ', @orig) . ') => (' . join(', ', @new) . ')');
}

sub pmx_log {
    my $msg = shift;
    my $args = shift;
    $msg->log($args->param(0)->[0], $args->param(1)->[0]);
}

sub pmx_relay {
    my $msg = shift;
    my $args = shift;

    # we can't query a multidimensional list unless we have envelope
    # data
    my $is_multilist = $args->is_multilist();
    $msg->assert_event('envelope') if $is_multilist;
    my $relay;
    if ($args->tag('fur')) {
        $relay = $msg->minfo_value("Fur");
        $relay .= ' '. $msg->minfo_value('Fur_rdns');
    }
    else {
        $relay = $msg->minfo_value("relay");
    }
    my @relay  = split / /, $relay;
    s/^\[(.*)\]$/$1/ foreach @relay;
    if ($args->is_multilist()) {
        return _str_compare($msg, $args, @relay);
    }
    else {
        return _str_compare(undef, $args, @relay);
    }
}

sub redirect {
    my $msg = shift;
    my $args = shift;
    $msg->assert_event('envelope');
    $msg->minfo_value("Envelope-To",@{$args->param(0)});
}

sub pmx_add_recipient {
    my $msg = shift;
    my $args = shift;
    $msg->assert_event('envelope');
    my @recips = $msg->minfo_value("Envelope-To");
    push @recips, @{$args->param(0)};
    $msg->minfo_value("Envelope-To", @recips);
}

sub pmx_change_from {
    my $msg = shift;
    my $args = shift;
    $msg->assert_event('envelope');
    $msg->minfo_value("Envelope-From", $args->param(0));
}

sub _handle_utf8_header {
    my $value = shift;
    # The header should be a mix of UTF-8 from the policy and 
    # encoded =?...?= blocks from the message (which are ascii and
    # therefore also UTF-8).  If this is not the case (ie. a raw Chinese
    # Chinese header), we can't do much at this stage because now we have
    # a half UTF-8, half other encoded string.
    unless (is_valid_utf8($value)) {
        warn "Header is not in UTF-8";
        return $value;
    }
    # convert any encoded parts (=?~?=) to raw utf8
    utf8_header($value, \%utf8_config);
    # now return the recoded header
    return encode_utf8_header($value);
}

sub pmx_add_header {
    my $msg = shift;
    my $args = shift;

    my $value;
    eval { $value = $args->param_decoded(1)->[0] };
    if ($@) {
        # param_decoded could not decode the string to latin1.
        # We may be dealing with parameter containing non-latin characters.
        $value = _handle_utf8_header($args->param(1)->[0]);
    }
    else {        
        recode_header($value) if($args->tag('recode'));
    }
    $msg->header_add($args->param(0)->[0], $value);
}

sub pmx_replace_header {
    my $msg = shift;
    my $args = shift;

    my $key = $args->param(0)->[0];
    my $value;
    eval { $value = $args->param_decoded(1)->[0] };
    if ($@) {
        # param_decoded could not decode the string to latin1.
        # We may be dealing with parameter containing non-latin characters.
        $value = _handle_utf8_header($args->param(1)->[0]);
    }
    else {
        recode_header($value) if($args->tag('recode'));
    }

    if($args->tag('all')) {
	$msg->header_replace($key, $value, -1);
    }
    else {
	my $index_tag = $args->tag('index');

	## Commenting it will avoid adding ':index 0' tags
	## for all existing scripts.
	## $msg->log('WARNING', 'pmx_replace_header: All or index tag is needed') unless($index_tag);

	my $index = $index_tag->{params}->[0]->{value} || 0;
	$msg->header_replace($key, $value, $index);
    }
}

sub pmx_delete_header {
    my $msg = shift;
    my $args = shift;

    my $key = $args->param(0)->[0];

    if($args->tag('all')) {
	$msg->header_delete($key, -1);
    }
    else {
	my $index_tag = $args->tag('index');

	## Commenting it will avoid adding ':index 0' tags
	## for all existing scripts.
	## $msg->log('WARNING', 'pmx_delete_header: All or index tag is needed') unless($index_tag);

	my $index = $index_tag->{params}->[0]->{value} || 0;
	$msg->header_delete($key, $index);
    }
}

sub pmx_deliver_for {
    my ($msg,$args) = @_;
    
    $msg->assert_event('eom');
        
    my @matching;
    my @non_matching;
    my @recips = $msg->minfo_value("Envelope-To");
    if (my $matcher = $args->matcher(scalar $msg->minfo_value('Envelope-To'))) {
        foreach my $recip (@recips) {
            if ($matcher->($recip)) {
                push @matching, $recip;
            }
            else {
                push @non_matching, $recip;
            }
        }
    }
    else {
        croak "no match function found!";
    }

    my ($optin,$optout);
    if ($args->tag('except')) {
        $optout = \@non_matching;
        $optin = \@matching;
    }
    else {
        $optout = \@matching;
        $optin = \@non_matching;
    }
    
    unless (@$optout) {
        $msg->log('DEBUG','pmx_deliver_for:  Continue processing for all recipients');
        return;
    }
    
    unless (@$optin) {
        $msg->log('DEBUG','pmx_deliver_for:  Deliver immediately for all recipients');
        $msg->delivery_action('keep');
        $msg->stop;
        return;
    }
    
    queue_mail(tempfile => $msg->implode_to_tempfile(), to=>$optout, minfo => $msg->minfo());
    $msg->minfo_value("Envelope-To", @$optin);
    $msg->log('DEBUG',
              'pmx_deliver_for: Continue processing for ('
              . join(', ', @$optin)
              . '), deliver immediately for ('
              . join(', ', @$optout) . ')');
}

sub pmx_mta_macro {
    my $msg   = shift;
    my $args  = shift;
    my $name  = $args->param(0)->[0];
    my $value = $args->param(1)->[0];

    if (grep {"macro_$name" eq $_} $msg->attrs) {
        return ($msg->get_attr("macro_$name") eq $value);
    }
    return 0;
}

sub pmx_mark {
    my $msg = shift;
    my $args = shift;
    $msg->assert_event('envelope');
    my $name = $args->param(0)->[0];
    my $value = $args->param(1)->[0];
    my $mname = (is_external_mark($name)) ? "x1mark_$name" : "mark_$name";
    $msg->set_attr($mname, $value);
}

sub pmx_mark1 {
    my $msg = shift;
    my $args = shift;
    $msg->assert_event('envelope');
    my $name = $args->param(0)->[0];
    my $mname = (is_external_mark($name)) ? "x1mark_$name" : "mark_$name";
    $msg->set_attr("$mname", undef);
    $msg->minfo_value(Direction => 'Outbound') if lc($name) eq 'i';
}

sub pmx_test_mark {
    my $msg = shift;
    my $args = shift;
    if (my $v = $msg->header("X-PMX-Test-Request")) {
	$msg->set_attr("mark_Test-ID", $v);
	$msg->header_add("X-PMX-Test-Response", $v);
	$msg->minfo_value(Relay => "fake.test.ActiveState.com");
    }
}

sub SUBJECT {
    my $msg = shift;
    my $s = $msg->header('Subject', -1);
    if (defined $s) {
        # remove all low bit control chars
        $s =~ tr/\x00-\x1F//d;
        return $s;
    } else {
        return '';
    }
}

sub MESSAGE_SIZE {
    my $msg = shift;
    return $msg->message_size;
}

sub BODY_SIZE {
    my $msg = shift;
    return $msg->body_raw_size;
}

sub HEADER_SIZE {
    my $msg = shift;
    return $msg->header_size;
}

sub QUEUE_ID {
    my $msg = shift;
    return $msg->minfo_value('Queue-Id');
}

sub SENDER_IP {
    my $msg = shift;
    my $relay = $msg->minfo_value('Relay');
    return unless defined $relay;
    my (undef,$ip) = $relay =~ /^(.*) \[(.*)\]/;
    return $ip || $relay;
}

sub DATETIME {
    my $msg = shift;
    return scalar localtime(time);
}

sub DATETIME_GMT {
    return scalar gmtime(time);
}

sub ENVELOPE_FROM {
    my $msg = shift;
    return $msg->minfo_value('Envelope-From');
}

sub ENVELOPE_TO {
    my $msg = shift;
    return join(', ', $msg->envelope_recipients);
}

sub HEADER_TO {
    my $msg = shift;
    return $msg->header('To', -1);
}

sub HEADER_FROM {
    my $msg = shift;
    return $msg->header('From', -1);
}

sub HEADER_CC {
    my $msg = shift;
    return $msg->header('Cc', -1);
}

sub HEADER_DATE {
    my $msg = shift;
    return $msg->header('Date', -1);
}

1;
