package PureMessage::Policy::Plugin::Utils;
use strict;
use PureMessage::Config qw(expand_tmpl_filename);
use PureMessage::Template qw(expand_template);
use PureMessage::Text::Wrap qw(wrap);

use MIME::QuotedPrint qw(encode_qp decode_qp);
use MIME::Base64 qw(encode_base64 decode_base64);

use PureMessage::MultifileCache;
use PureMessage::Version qw($PREFIX);

use base 'Exporter';
our @EXPORT_OK = qw(set_attachments
                    get_attachments
                    set_attachment_filter_types
                    get_attachment_filter_types
                    set_attachment_filter_filetypes
                    get_attachment_filter_filetypes
                    set_attachment_filter_extensions
                    get_attachment_filter_extensions
                    get_recipient
                    get_recipients
                    get_matcher
                    expand_localized_tmpl
                    expand_part_tmpl
                    recode_header
                    find_multifiles
                    normalize_addrs
                    replace_nested_part
                    get_action_data
                    );

my $multifile_cache = undef;
eval { $multifile_cache = PureMessage::MultifileCache->new(); };

sub set_attachments {
    my ($msg, $matching, $non_matching) = @_;
    my $a = $msg->get_attr('pmxpe_attachment_filter');
    
    my @new;
    if ($msg->context_flip) {
        @new = @$non_matching;
    }
    else {
        @new = @$matching;
    }

    my @save;
    if ($a and @$a) {
        my $top = $a->[$#$a];
        if ($top and @$top) {
            foreach my $id (@new) {
                push @save, $id if grep $_ eq $id, @$top;
            }
        }
        else {
            @save = @new;
        }
    }
    else {
        @save = @new;
    }

    _set_context_templates($msg, @save);
    
    push @$a, \@save;
    $msg->set_attr('pmxpe_attachment_filter', $a, 1);
    
    return $msg->context_flip ? (@$non_matching == 0) : (@$matching != 0);
}

sub _set_context_templates {
    my ($msg, @parts) = @_;

    my $value = _get_attachment_names($msg, @parts);
    my $attr = _set_attr($msg, 'pmxpe_attachment_names', $value);
    $msg->{template}->template_context('ATTACHMENT_NAMES', $attr);
}

sub set_attachment_filter_types {
    my ($msg, $args) = @_;
    my $a = $msg->get_attr('pmxpe_attachment_filter_types') || [ ];
    # Get the array ref of mimetypes set in pmx_attachment_type
    my $attachment_filters = $args->{'p'}->[0]->{'value'} || [ ];
    if ($msg->context_flip) {
        $attachment_filters = [ ];
    }
    push @$a, $attachment_filters;

    $msg->set_attr('pmxpe_attachment_filter_types', $a, 1);
    
    return $msg->context_flip ? 1 : (@$a != 0);
}

sub _get_attachment_names {
    my ($msg, @parts) = @_;

    return 'null' unless @parts;

    my @names;
    $msg->enumerate_parts(sub {
        my $part = shift;
        my $id = $part->part_id;

        foreach (@parts) {
            if($_ eq $id) {
                my (undef, %ct_p) = $part->content_type;
		my $cd = $part->header_decoded("Content-Disposition");
		my $cd_name = $1
		    if $cd and $cd =~ /filename\s*=\s*\"?([^\r\n;"]*)/i;

		if(defined $ct_p{name} and defined $cd_name) {
		    push @names, ($ct_p{name} eq $cd_name)
			? "$ct_p{name}"
			: "[$ct_p{name}, $cd_name]";
		}
		else {
		    push @names, "$ct_p{name}" if defined $ct_p{name};
		    push @names, "$cd_name" if defined $cd_name;
		}
            }
        }
    });

    return @names ? join(', ', @names) : 'null';
}

sub _set_attr {
    my ($msg, $name, $value) = @_;

    return unless (defined $name and defined $value);

    my $attr = $msg->get_attr($name);
    push @$attr, $value;
    $msg->set_attr($name, $attr, 1);

    return $msg->get_attr($name);
}

sub get_attachments {
    my $msg = shift;
    my $a = $msg->get_attr('pmxpe_attachment_filter');
    my @attach = ();
    if ($a and @$a) {
        @attach = @{$a->[$#$a]};
    }
    return @attach;
}

sub get_attachment_filter_types {
    my $msg = shift;
    my $a = $msg->get_attr('pmxpe_attachment_filter_types');
    my @attach = ();
    if ($a and @$a) {
        @attach = @{$a->[$#$a]};
    }
    return @attach;
}

sub get_attachment_filter_extensions {
    my $msg = shift;
    my $a = $msg->get_attr('pmxpe_attachment_filter_extensions');
    my @attach = ();
    if ($a and @$a) {
        @attach = @{$a->[$#$a]};
    }
    return @attach;
}

sub set_attachment_filter_extensions {
    my ($msg, $args) = @_;
    my $a = $msg->get_attr('pmxpe_attachment_filter_extensions') || [ ];
    # Get the array ref of mimetypes set in pmx_attachment_type
    my $attachment_filters = $args->{'p'}->[0]->{'value'} || [ ];
    if ($msg->context_flip) {
        $attachment_filters = [ ];
    }
    push @$a, $attachment_filters;

    $msg->set_attr('pmxpe_attachment_filter_extensions', $a, 1);
    
    return $msg->context_flip ? 1 : (@$a != 0);
}

sub get_attachment_filter_filetypes {
    my $msg = shift;
    my $a = $msg->get_attr('pmxpe_attachment_filter_filetypes');
    my @attach = ();
    if ($a and @$a) {
        @attach = @{$a->[$#$a]};
    }
    return @attach;
}

sub set_attachment_filter_filetypes {
    my ($msg, $args) = @_;
    my $a = $msg->get_attr('pmxpe_attachment_filter_filetypes') || [ ];
    # Get the array ref of mimetypes set in pmx_attachment_type
    my $attachment_filters = $args->{'p'}->[0]->{'value'} || [ ];
    if ($msg->context_flip) {
        $attachment_filters = [ ];
    }
    push @$a, $attachment_filters;

    $msg->set_attr('pmxpe_attachment_filter_filetypes', $a, 1);
    
    return $msg->context_flip ? 1 : (@$a != 0);
}

sub get_recipients {
    my $msg = shift;
    
    my @recips = $msg->mapped_recipients;
    if (@recips and defined $recips[0]) {
        return wantarray ? @recips : $recips[0];
    }
    else {
        # backward compatibility for old PM-Milter-Policy
        $msg->log('WARNING', "POLICY WARNING: No mapped recipient(s) found."
        . "  Either PureMessage-Milter-Policy is old or there is a bad map entry"
        . " in the recipient-aliases map.  Using original recipients instead.");
        return $msg->envelope_recipients;
    }
}

sub get_recipient {
    return scalar get_recipients(@_);
}

sub normalize_addrs {
    my @addrs = @_;

    for (@addrs) {
        s/\+[^@]*//; # strip detail/subaddressing
    }
    return @addrs;
}

sub _no_match {
    return 0;
}

sub get_matcher {
    my $msg = shift;
    my $args = shift;
    my $list_name = shift;

    my $key = get_recipient($msg);
    if($args->{lm} && $args->{lists} && $args->is_multilist) {
        my $_list_name = $list_name ? $list_name : $args->{lists}[0];
        return \&_no_match unless $_list_name;
        my $mlist = $args->{lm}->find_list($_list_name, want_ml => 1);
        if($msg && $mlist && $mlist->isa('PureMessage::MultiList::multigroupfile')) {
            if($msg->minfo_value('Direction') eq 'Outbound') {
                my $group = $msg->minfo_value('Sender-Group');
                return \&_no_match unless $group;
                ($key) = $group =~ /\|([^|]+)$/;
            }
            else {
                my @groups = $msg->minfo_value('Recipient-Group');
                return \&_no_match unless @groups;
                my ($match) = grep /\Q$key\E/, @groups;
                return \&_no_match unless $match;
                ($key) = $match =~ /\|([^|]+)$/;
            }
            return \&_no_match unless $key;
        }
    }

    # if listname parameter is given, return the corresponding list
    if ($list_name) {
        return $args->find_list($list_name, key=>$key)->matcher;
    }
    # return the default list
    return $args->matcher($key);
}

sub expand_localized_tmpl {
    my $msg = shift;
    my $tmpl = shift;
    my $lang = $msg->minfo_value("User-Language");
    return expand_tmpl_filename($tmpl, $lang);
}

sub expand_part_tmpl {
    my $part = shift;
    my $tmpl = shift;
    my $vars = $part->get_info("part_vars");

    my ($ct, %ct_p) = $part->content_type;
    my $cd = $part->header_decoded("Content-Disposition") || "inline";
    my @names;
    push @names, $ct_p{name} if exists $ct_p{name};
    if($cd and $cd =~ /filename\s*=\s*\"?([^\r\n;"]*)/i) {
		push @names, $1 unless exists $ct_p{name} and $1 eq $ct_p{name};
    }
    push @names, "none ($cd)" unless @names;

    return expand_template($tmpl, vars => {
	attachment_name => join(", ", @names),
	attachment_type => $ct,
	attachment_size => $part->body_raw_size . " bytes",
	$vars ? %$vars : (),
    });
}

my $ENCODED_WORD = qr/^(.+?)=\?([^?]+)\?([^?]+)\?([^?]+)\?=/s;

sub recode_header {
    return unless (defined $_[0]) and ($_[0] =~ $ENCODED_WORD);

    my ($leading_text, $charset, $encoding, $encoded_text) =
	($1, $2, lc($3), $4);
    my ($begin, $end) = ($-[0], $+[0]);

    my ($encoder, $decoder);

    if($encoding eq 'b') {
	$encoder = \&encode_base64;
	$decoder = \&decode_base64;
    }
    elsif($encoding eq 'q') {
	$encoder = \&encode_qp;
	$decoder = \&decode_qp;
    }
    else { return; }

    $leading_text =~ s/[\r\n\t]+//g;

    $encoded_text = $decoder->($encoded_text);
    $encoded_text = $encoder->("$leading_text$encoded_text", '');

    # $_[0] =~ s/$ENCODED_WORD/=\?$charset\?$encoding\?$encoded_text\?=/;
    substr($_[0], $begin, $end - $begin) =
	"=?$charset?$encoding?$encoded_text?=";
}

sub find_multifiles {
    my ($multifiles) = @_;
    return unless $multifiles;

    while(my ($id, $charset) = each %$multifiles) {
        my $multifile = $multifile_cache->get_multifile($id);
        if($charset) {
            die "Unmatched charset for multifile '$id': $charset" unless
                $multifile->get_charset =~ /$charset/;
        }
    }
}

sub replace_nested_part {
    my ($msg, $part_id, $repl, $callback, %opts) = @_;

    # Make sure that the body_filename exists. Overzealous virus cleaning
    # backends can delete it, which throws a warning.
    my $part_fetch = $msg->part_nested($part_id, 1)
	    or die "can't fetch part '$part_id'";
    unless (-f (my $fname = $part_fetch->body_filename)) {
        open (my $TOUCH, '>', $fname)
            or die "can't recreate part ${part_id} body in $fname: $!";
    }

    open(my $TEMPLATE, $repl) or die "can't read $repl: $!";

    # Read message directives and headers
    my (%directive, @hdr);
    while (<$TEMPLATE>) {
        last if /^$/;
        my ($fname, $fval) = /^([\w-]+)\s*:\s*(.*)/;
        die "Bad template header in '$repl': $_" unless $fname;
        if ($fname =~ s/X-PMX-//i) {
            $directive{lc $fname} = $fval;
        }
        else {
            chomp($fname, $fval);
            push @hdr, $fname, $fval;
        }
    }

    # Read the template body
    my $template = do { local $/; <$TEMPLATE> };
    close $TEMPLATE;

    # Expand the templates
    for my $tmpl ($template, @hdr) {
        $tmpl = $callback->($msg, $tmpl, $part_id);
    }

    # Apply the directives
    if (exists $directive{wrap}) {
    	$template = wrap($template, $directive{wrap});
    }

    # Replace the part
    if ($part_id eq '1') {

        if ($opts{warn_only}) {
            my ($tmp_msg) = $msg->make_multipart(prepend=>1);
            unless (defined $tmp_msg) {
                my $ct = $msg->content_type;
                $msg->log("NOTICE", "POLICY WARNING: unable to add warning banner to '$ct' part");
                return;
            }
            $msg = $tmp_msg;
        }

        while (@hdr) {
            my ($h, $v) = (shift(@hdr), shift(@hdr));
            $msg->header_replace($h, $v);
        }
        $msg->body_copy_from($template);
    }
    else {
        my ($parent, $idx) = do {
            my (@p, $p, $idx) = split /\./, $part_id;
            $idx = pop @p;
            $p = join('.', @p);
            ($p, $idx);
        };

        my $ppart = $msg->part_nested($parent, 1)
            or die "can't fetch part '$parent' while replacing '$part_id'";

        my $part;
        if ($opts{warn_only}) {
            $part = $ppart->part("1.$idx");
            my ($tmp_part) = $part->make_multipart(prepend=>1);

            unless (defined $tmp_part) {
                my $ct = $part->content_type;
                $msg->log("NOTICE", "POLICY WARNING: unable to add warning banner to '$ct' part");
                return;
            }
	        $part = $tmp_part;
        }
        else {
            $part = $ppart->part_replace("1.$idx");
        }

        while (@hdr) {
            my ($h, $v) = (shift(@hdr), shift(@hdr));
            $part->header_add($h, $v);
        }
        $part->body_copy_from($template);
    }
}

sub _get_groups {
    my ($msg, $args) = @_;
    my @groups = ();

    my $direction = $msg->minfo_value('Direction');
    my @_groups = ($direction eq 'Outbound')
        ? $msg->minfo_value('Sender-Group')
        : $msg->minfo_value('Recipient-Group');

    if(@_groups) {
        my %_groups = ();
        for my $_group (@_groups) {
            my ($recipient, $group) = split /\|/, $_group;
            # After message splitting, re-map mapped-recipients (part of the
            # entire list of recipient) to their respective groups
            $_groups{$recipient} = $group;
        }
        if($direction eq 'Outbound') {
            push @groups, (values %_groups)[0];
        }
        else {
            for my $recipient (get_recipients($msg)) {
                push @groups, $_groups{$recipient} if $_groups{$recipient};
            }
        }
    }

    return \@groups;
}

sub get_action_data {
    my ($msg,$args) = @_;
    my $data = "";

    # Fetch group-specific data first
    my $group_tag = $args->tag('groupfile');
    if($multifile_cache && $group_tag) {
        my $groups = _get_groups($msg, $args);
        if($groups && @$groups) {
            eval {
                my $multifile_id = $group_tag->{params}->[0]->{value}->[0];
                my $multifile = $multifile_cache->get_multifile($multifile_id);
                # Only use the first group because, once we hit this,
                # it means the message has been split and we only have
                # one group.
                $data = $multifile->get_encoded_content($groups->[0]);
            };
            return $data unless $@;
        }
    }

    if ($args->tag('data')){
        $data = $args->param(0)->[0];
    }
    else {
        my $file = $args->param(0)->[0];
        # 0+ as filename does not make sense
        return undef unless $file;
        
        #relative paths are interpreted relative to PREFIX
        $file = "$PREFIX/$file" unless index($file, '/') == 0;
        if (open F, $file) {
            my $str = join '', <F>;
            close F;
            $data = $msg->template_expand($str);
        }
        else {
            $msg->log("ERR", "POLICY ERROR: could not open '$file': $!");
        }
    }
    return $data;
}

__END__

=head1 NAME

PureMessage::Policy::Plugin::Utils - helper functions for plugins

=head1 DESCRIPTION

=head2 Attachment Filtering

Provides two functions help implement attachment filtering.  Attachment
filtering is implemented by maintaining a stack of arrays of attachments.  The
array at the top of the stack is the "current" set of attachments that is
operated on.

The message object provides a facility for contextual attributes.  If the
attribute is set with the "is_contextual" flag, the message object will assume
that the attribute value is a stack that should be popped when the policy
engine returns from a context (the message object gets this signal from the
policy engine).

=head2 Getting Mapped Recipients

The get_recipients call provides a level of indirection for retrieving the
mapped recipients of a message.  It encapsulates error handling logic for the
case where the mapped-recipients field is empty.

=head2 Functions

=over 4

=item B<get_attachments($msg)>

Get the current list of attachments.  Returns a list of part ids.

=item B<set_attachments($msg,\@matching_attachments, \@non_matching_attachments)>

Set the current list of attachments.  This function will decide which list
(matching or non-matching) to set based on the context of the policy engine
(whether the 'not' test is in effect).  If there is no current list already,
this becomes the current list.  If a current list does exist, the resulting
list is the intersection of the new list with the existing one.  This allows
for attachment filters to be chained together.

The return value of this function ought to be used to be the return value of a
policy test that implements attachment filtering.  It is calculated as follows:

                     matching list chosen  non-matching list chosen

 list is empty               FALSE                  TRUE
 
 list is non-empty           TRUE                   FALSE

=item B<get_recipients($msg)>

Return the message's current mapped recipients.  If the mapped-recipients
field is empty, return the envelope recipients instead.

=item B<get_recipient($msg)>

Same as get_recipients(), except only the first recipient is returned.

=item B<normalize_addrs(@addr_list)>

Returns list of clean email addresses, this currently involves removing the detail 
component of each address.

=item B<get_matcher($msg, $args)>

Returns a matcher object from the plugin's $args.  Encapsulates end-user list
lookups using mapped recipients.

=item B<get_matcher($msg, $args, $listname)>

Specify a list name to get a matcher from a specific list.

=item B<expand_localized_tmpl($msg, $tmpl_filename)>

Expand a template filename using user language preference retrieved from the message.
Returns the expanded filename.

=item B<expand_part_tmpl($part, $template)>

Expand part-specific template variables in B<$template>.

=item B<recode_header(\$hdr)>

Re-code headers if there is plain text in front of encoded-word(s).

It targets to some MUAs that cannot display 'plain-text + encoded-word + ...'
correctly. recode_header() re-codes the plain-text and the first encoded-word
into a new encoded-word.

=back

=head1 COPYRIGHT

Copyright (C) 2000-2016 Sophos Group. All rights reserved.
